//=============================================================================
// UdpServerQuery
//
// Author: Brandon "GreenMarine" Reinhart
// Version: 1.0
//
// This query server is compliant with the GameSpy Uplink Specification.
// The specification is available at http://www.gamespy.com/developer
// and might be of use to progammers who are writing or maintaining
// their own stat gathering/game querying software.
//
// Note: Currently, SendText returns false if successful.
//
// Full documentation on this class is available at http://unreal.epicgames.com/
//
//=============================================================================
class UdpServerQuery expands UdpLink config;

// Game Server Config.
var() config string[128]	ServerName;			// Name of the server, i.e.: Bob's Server.
var() config string[128] 	AdminName;			// Name of the server admin.
var() config string[128] 	AdminEmail;			// Email address of the server admin.
var() config int 			Region;				// Region of the game server.
var() name					QueryName;			// Name to set this object's Tag to.
var int					    CurrentQueryNum;	// Query ID Number.

// Initialize.
function PreBeginPlay()
{
	// Set the Tag
	Tag = QueryName;

	// Bind the listen socket
	if( !BindPort(GetServerPort(),true) )
	{
		Log("UdpServerQuery: Port failed to bind.");
		return;
	}
	Log("UdpServerQuery: Port "$Port$" succesfully bound.");
}

// Return the server's port number.
function int GetServerPort()
{
	local string[64] S;
	local int i;

	// Figure out the server's port.
	S = Level.GetAddressURL();
	i = InStr( S, ":" );
	assert(i>=0);
	return int(Mid(S,i+1));
}

// Received a query request.
event ReceivedText(IpAddr Addr, string[240] Text)
{
	local string[240] Query;
	local bool QueryRemaining;
	local int  QueryNum, PacketNum;

	// Assign this packet a unique value from 1 to 100
	CurrentQueryNum++;
	if (CurrentQueryNum > 100)
		CurrentQueryNum = 1;
	QueryNum = CurrentQueryNum;

	Query = Text;
	if (Query == "")		// If the string is empty, don't parse it
		QueryRemaining = false;
	else
		QueryRemaining = true;
	
	while (QueryRemaining) {
		Query = ParseQuery(Addr, Query, QueryNum, PacketNum);
		if (Query == "")
			QueryRemaining = false;
		else
			QueryRemaining = true;
	}
}

function string[240] ParseQuery(IpAddr Addr, string[240] Query, int QueryNum, out int PacketNum)
{
	local string[240] QueryType, QueryValue, QueryRest, ValidationString;
	local bool Result;
	local string[32] FinalPacket;
	
	Result = ParseNextQuery(Query, QueryType, QueryValue, QueryRest, FinalPacket);
	if( !Result )
		return "";

	switch( QueryType )
	{
		case "basic":
			Result = SendQueryPacket(Addr, GetBasic(), QueryNum, ++PacketNum, FinalPacket);
			break;
		case "info":
			Result = SendQueryPacket(Addr, GetInfo(), QueryNum, ++PacketNum, FinalPacket);
			break;
		case "rules":
			Result = SendQueryPacket(Addr, GetRules(), QueryNum, ++PacketNum, FinalPacket);
			break;
		case "players":
			if( DeathMatchGame(Level.Game)!=None && DeathMatchGame(Level.Game).NumPlayers > 0 )
				Result = SendPlayers(Addr, QueryNum, PacketNum, FinalPacket);
			else
				Result = SendQueryPacket(Addr, "", QueryNum, PacketNum, FinalPacket);
			break;
		case "status":
			Result = SendQueryPacket(Addr, GetBasic(), QueryNum, ++PacketNum, "");
			Result = SendQueryPacket(Addr, GetInfo(), QueryNum, ++PacketNum, "");
			if( DeathMatchGame(Level.Game)!=None && DeathMatchGame(Level.Game).NumPlayers == 0 )
			{
				Result = SendQueryPacket(Addr, GetRules(), QueryNum, ++PacketNum, FinalPacket);
			}
			else
			{
				Result = SendQueryPacket(Addr, GetRules(), QueryNum, ++PacketNum, "");
				Result = SendPlayers(Addr, QueryNum, PacketNum, FinalPacket);
			}
			break;
		case "echo":
			// Respond to an echo with the same string
			Result = SendQueryPacket(Addr, "\\echo\\"$QueryValue, QueryNum, ++PacketNum, FinalPacket);
			break;
		case "secure":
			ValidationString = "\\validate\\"$Validate(QueryValue);
			Result = SendQueryPacket(Addr, ValidationString, QueryNum, ++PacketNum, FinalPacket);
			break;
		case "level_property":
			Result = SendQueryPacket(Addr, GetLevelProperty(QueryValue), QueryNum, ++PacketNum, FinalPacket);
			break;
		case "game_property":
			Result = SendQueryPacket(Addr, GetGameProperty(QueryValue), QueryNum, ++PacketNum, FinalPacket);
			break;
		case "player_property":
			Result = SendQueryPacket(Addr, GetPlayerProperty(QueryValue), QueryNum, ++PacketNum, FinalPacket);
			break;
		default:
			break;
	}
	if( Result )
		Log("UdpServerQuery: Error responding to query.");
	return QueryRest;
}

// SendQueryPacket is a wrapper for SendText that allows for packet numbering.
function bool SendQueryPacket(IpAddr Addr, string[240] SendString, int QueryNum, int PacketNum, string[32] FinalPacket)
{
	local bool Result;
	if (FinalPacket == "final") {
		SendString = SendString$"\\final\\";
	}
	SendString = SendString$"\\queryid\\"$QueryNum$"."$PacketNum;

	Result = SendText(Addr, SendString);

	return Result;
}

// Return a string of basic information.
function string[240] GetBasic() {
	local string[240] ResultSet;

	// The name of this game.
	ResultSet = "\\gamename\\unreal";

	// The version of this game.
	ResultSet = ResultSet$"\\gamever\\"$Level.EngineVersion;

	// The regional location of this game.
	ResultSet = ResultSet$"\\location\\"$Region;
	
	return ResultSet;
}

// Return a string of important system information.
function string[240] GetInfo() {
	local string[240] ResultSet;
	
	// The server name, i.e.: Bobs Server
	ResultSet = "\\hostname\\"$ServerName;

	// The server port.
	ResultSet = ResultSet$"\\hostport\\"$GetServerPort();

	// (optional) The server IP
	// if (ServerIP != "")
	//	ResultSet = ResultSet$"\\hostip\\"$ServerIP;

	// The map/level name
	ResultSet = ResultSet$"\\mapname\\"$Level.Title;

	// The mod or game type
	ResultSet = ResultSet$"\\gametype\\"$Level.Game.Class;

	if( DeathMatchGame(Level.Game)!=None )
	{
		// The number of players
		ResultSet = ResultSet$"\\numplayers\\"$DeathMatchGame(Level.Game).NumPlayers;

		// The maximum number of players
		ResultSet = ResultSet$"\\maxplayers\\"$DeathMatchGame(Level.Game).MaxPlayers;
	}

	// The game mode: openplaying
	ResultSet = ResultSet$"\\gamemode\\openplaying";

	return ResultSet;
}

// Return a string of miscellaneous information.
// Game specific information, user defined data, custom parameters for the command line.
function string[240] GetRules()
{
	local string[240] ResultSet;
	local DeathmatchGame DM;
	DM = DeathmatchGame(Level.Game);
	if( DM!=None )
	{
		// Timelimit.
		ResultSet = "\\timelimit\\"$DM.TimeLimit;
		
		// Fraglimit
		ResultSet = ResultSet$"\\fraglimit\\"$DM.FragLimit;
		
		// Bots in Multiplay?
		if( DM.bMultiplayerBots)
			Resultset = ResultSet$"\\MultiplayerBots\\"$true;
		else
			Resultset = ResultSet$"\\MultiplayerBots\\"$false;

		// Change levels?
		if( DM.bChangeLevels )
			Resultset = ResultSet$"\\ChangeLevels\\"$true;
		else
			Resultset = ResultSet$"\\ChangeLevels\\"$false;
	}

	// Admin's Name
	if( AdminName != "" )
		ResultSet = ResultSet$"\\AdminName\\"$AdminName;
	
	// Admin's Email
	if( AdminEmail != "" )
		ResultSet = ResultSet$"\\AdminEMail\\"$AdminEmail;

	return ResultSet;
}

// Return a string of information on a player.
function string[240] GetPlayer( PlayerPawn P, int PlayerNum )
{
	local string[240] ResultSet;
	
	// Name
	ResultSet = "\\player_"$PlayerNum$"\\"$P.PlayerName;

	// Frags
	ResultSet = ResultSet$"\\frags_"$PlayerNum$"\\"$int(P.Score);
	
	// Ping
	ResultSet = ResultSet$"\\ping_"$PlayerNum$"\\"$P.ConsoleCommandResult("GETPING");
	
	// Team
	ResultSet = ResultSet$"\\team_"$PlayerNum$"\\"$P.Team;
	
	// Skin
	ResultSet = ResultSet$"\\skin_"$PlayerNum$"\\"$P.Skin;
	
	// Mesh
	ResultSet = ResultSet$"\\mesh_"$PlayerNum$"\\"$P.Mesh;
	
	// Time Playing

	return ResultSet;
}

// Send data for each player
function bool SendPlayers(IpAddr Addr, int QueryNum, out int PacketNum, string[32] FinalPacket)
{
	local Pawn P;
	local int i;
	local bool Result, SendResult;
	local DeathMatchGame DM;
	
	Result = false;

	P = Level.PawnList;
	DM = DeathMatchGame(Level.Game);
	if( DM!=None )
	{
		while( i < DM.NumPlayers )
		{
			if (P.IsA('PlayerPawn'))
			{
				if( i==DM.NumPlayers-1 && FinalPacket=="final" )
					SendResult = SendQueryPacket(Addr, GetPlayer(PlayerPawn(P), i), QueryNum, ++PacketNum, "final");
				else
					SendResult = SendQueryPacket(Addr, GetPlayer(PlayerPawn(P), i), QueryNum, ++PacketNum, "");
				Result = SendResult || Result;
				i++;
			}
			P = P.nextPawn;
		}
	}

	return Result;
}

// Get an arbitrary property from the level object.
function string[240] GetLevelProperty(string[240] Prop)
{
	local string[240] ResultSet;
	
	ResultSet = "\\"$Prop$"\\"$Level.GetPropertyText(Prop);
	
	return ResultSet;
}

// Get an arbitrary property from the game object.
function string[240] GetGameProperty(string[240] Prop)
{
	local string[240] ResultSet;

	ResultSet = "\\"$Prop$"\\"$Level.Game.GetPropertyText(Prop);
	
	return ResultSet;
}

// Get an arbitrary property from the players.
function string[240] GetPlayerProperty(string[240] Prop)
{
	local string[240] ResultSet;
	local int i;
	local PlayerPawn P;

	foreach AllActors(class'PlayerPawn', P) {
		i++;
		ResultSet = ResultSet$"\\"$Prop$"_"$i$"\\"$P.GetPropertyText(Prop);
	}
	
	return ResultSet;
}

defaultproperties
{
     ServerName="Another Unreal Server"
     QueryName=MasterUplink
}
